"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const Debugger = require("debug");
const methods_1 = require("./methods");
const query_1 = require("./query");
const debug = Debugger('akita:model');
class Model {
    constructor(data, params) {
        if (data) {
            Object.assign(this, data);
        }
        if (params) {
            Object.defineProperty(this, '__params', {
                value: Object.assign({}, params),
                writable: true,
                enumerable: false,
                configurable: true
            });
        }
    }
    static create(data) {
        let query = new query_1.default(this, 'create');
        query._data = data || {};
        return query;
    }
    static update(id, data) {
        let query = new query_1.default(this, 'update');
        if (data) {
            query._data = data;
        }
        if (id && data) {
            query._id = id;
            query._data = data;
        }
        else if (typeof id === 'object') {
            query._data = id;
        }
        return query;
    }
    static remove(conditions) {
        let query = new query_1.default(this, 'remove');
        if (conditions !== null && typeof conditions === 'object') {
            query.where(conditions);
        }
        else if (conditions !== undefined) {
            query._id = conditions;
        }
        return query;
    }
    static count(conditions) {
        let query = new query_1.default(this, 'count');
        if (conditions) {
            query.where(conditions);
        }
        return query;
    }
    static watch(conditions) {
        let query = new query_1.default(this, 'watch');
        if (conditions) {
            query.where(conditions);
        }
        return query;
    }
    static find(conditions) {
        let query = new query_1.default(this, 'find');
        if (conditions) {
            query.where(conditions);
        }
        return query;
    }
    static paginate(conditions) {
        let query = new query_1.default(this, 'paginate');
        if (conditions) {
            query.where(conditions);
        }
        return query;
    }
    static findOne(conditions) {
        let query = new query_1.default(this, 'findOne');
        query.limit(1);
        if (conditions) {
            if (typeof conditions === 'object') {
                query.where(conditions);
            }
            else {
                throw new Error('Akita Error: invalid params for Query#findOne()');
            }
        }
        return query;
    }
    static findByPk(id) {
        let query = new query_1.default(this, 'findByPk');
        query._id = id;
        return query;
    }
    static request(path, init, query, reducer) {
        let p = this.path || '';
        if (!p.endsWith('/') && path) {
            p += '/';
        }
        if (path.startsWith('/')) {
            path = path.substr(1);
        }
        path = p + path;
        if (path && init && (init.query || init.body)) {
            let queryParams = init.query && Object.assign({}, init.query);
            let bodyParams = init.body && (typeof init.body === 'object') && Object.assign({}, init.body);
            let params = {};
            let queryEmits = [];
            let bodyEmits = [];
            path = path.replace(/:\w+/g, (match) => {
                let key = match.substr(1);
                if (queryParams && queryParams.hasOwnProperty(key)) {
                    queryEmits.push(key);
                    params[key] = queryParams[key];
                    return encodeURIComponent(queryParams[key]);
                }
                if (bodyParams && bodyParams.hasOwnProperty(key)) {
                    bodyEmits.push(key);
                    params[key] = bodyParams[key];
                    return encodeURIComponent(bodyParams[key]);
                }
                return match;
            });
            if (query && (queryEmits.length || bodyEmits.length)) {
                query._params = params;
            }
            if (queryParams && queryEmits.length) {
                queryEmits.forEach((key) => {
                    delete queryParams[key];
                });
                init = Object.assign({}, init, { query: queryParams });
            }
            if (bodyParams && bodyParams.length) {
                bodyEmits.forEach((key) => {
                    delete bodyParams[key];
                });
                init = Object.assign({}, init, { body: bodyParams });
            }
        }
        return this.client.request(path, init, query, reducer);
    }
    request(path, init, reducer) {
        const M = this.constructor;
        const pk = M.pk || 'id';
        let id = this[pk];
        if (!id) {
            let method = (init && init.method) || 'GET';
            throw new Error(`Can not get pk field (${pk}) for ${method}: '${path}'`);
        }
        id = encodeURIComponent(id);
        if (!path || path.startsWith('/')) {
            path = id + path;
        }
        else {
            path = `${id}/${path}`;
        }
        let fullPath = `${M.path}/${path}`;
        let matchs = fullPath.match(/:\w+/g);
        if (matchs) {
            init = Object.assign({}, init);
            let query = Object.assign({}, init.query);
            matchs.forEach((match) => {
                let key = match.substr(1);
                if (!query.hasOwnProperty(key)) {
                    if (this.__params && this.__params.hasOwnProperty(key)) {
                        query[key] = this.__params[key];
                    }
                    else if (this.hasOwnProperty(key)) {
                        query[key] = this[key];
                    }
                    else if (this.hasOwnProperty(key)) {
                        query[key] = this[key];
                    }
                }
            });
            init.query = query;
        }
        return M.request(path, init, null, reducer);
    }
    save(init) {
        return this.request('', Object.assign({}, {
            method: 'PATCH',
            body: this.toJSON()
        }, init), () => undefined);
    }
    remove(init) {
        return this.request('', Object.assign({}, {
            method: 'DELETE'
        }, init), (json) => (json && json.removed) || 0);
    }
    toJSON() {
        let json = {};
        for (let key in this) {
            if (this.hasOwnProperty(key)) {
                json[key] = this[key];
            }
        }
        return json;
    }
}
exports.default = Model;
methods_1.default.forEach((method) => {
    Model[method] = function (path, init) {
        debug(`${this.name}.${method}`, path, init || '');
        init = init || {};
        init.method = method.toUpperCase();
        return this.request(path, init, null);
    };
    Model.prototype[method] = function (path, init) {
        const M = this.constructor;
        debug(`${M.name}.prototype.${method}`, path, init || '');
        init = init || {};
        init.method = method.toUpperCase();
        return this.request(path, init);
    };
});
