"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const Debugger = require("debug");
const debug = Debugger('akita:query');
class Query {
    constructor(model, op) {
        this[Symbol.toStringTag] = 'Promise';
        this.model = model;
        this._filters = null;
        this._data = null;
        this._page = 0;
        this._limit = 0;
        this._sort = '';
        this._id = null;
        this._args = null;
        this._search = '';
        this._op = op;
        this._promise = null;
        this._lastField = '';
    }
    arg(args, value) {
        if (this._result)
            throw new Error('Can not change query after exec.');
        if (!this._args) {
            this._args = {};
        }
        if (typeof args === 'object') {
            Object.assign(this._args, args);
        }
        else {
            this._args[args] = value;
        }
        return this;
    }
    where(conditions, value) {
        if (this._result)
            throw new Error('Can not change query after exec.');
        if (!this._filters) {
            this._filters = {};
        }
        if (typeof conditions === 'object' && value === undefined) {
            this._filters = Object.assign(this._filters, conditions);
        }
        else if (typeof conditions === 'string') {
            if (value === undefined) {
                this._lastField = conditions;
            }
            else {
                this._filters[conditions] = value;
            }
        }
        else {
            throw new Error('Akita Error: invalid params for Query#where()');
        }
        return this;
    }
    search(keyword) {
        if (this._result)
            throw new Error('Can not change query after exec.');
        this._search = keyword;
        return this;
    }
    __filter(type, value) {
        if (this._result)
            throw new Error('Can not change query after exec.');
        if (!this._filters) {
            this._filters = {};
        }
        if (!this._lastField) {
            throw new Error(`Akita Error: you should invoke .where(string) before .${type}()`);
        }
        if (type === 'eq') {
            this._filters[this._lastField] = value;
        }
        else {
            if (typeof this._filters[this._lastField] !== 'object') {
                this._filters[this._lastField] = {};
            }
            this._filters[this._lastField][`$${type}`] = value;
        }
        return this;
    }
    eq(value) {
        return this.__filter('eq', value);
    }
    ne(value) {
        return this.__filter('ne', value);
    }
    regex(value) {
        return this.__filter('regex', value);
    }
    in(value) {
        return this.__filter('in', value);
    }
    nin(value) {
        return this.__filter('nin', value);
    }
    lt(value) {
        return this.__filter('lt', value);
    }
    lte(value) {
        return this.__filter('lte', value);
    }
    gt(value) {
        return this.__filter('gt', value);
    }
    gte(value) {
        return this.__filter('gte', value);
    }
    limit(size) {
        if (this._result)
            throw new Error('Can not change query after exec.');
        this._limit = size;
        return this;
    }
    page(value) {
        if (this._result)
            throw new Error('Can not change query after exec.');
        this._page = value;
        return this;
    }
    sort(value) {
        if (this._result)
            throw new Error('Can not change query after exec.');
        this._sort = value;
        return this;
    }
    exec() {
        if (this._result) {
            return this._result;
        }
        this._debug();
        let init = this._createInit();
        let path = init.path;
        delete init.path;
        let reducer;
        const M = this.model;
        const createRecord = (data) => new M(data, this._params);
        switch (this._op) {
            case 'remove':
                reducer = (json) => (json && json.removed) || 0;
                break;
            case 'findOne':
                reducer = (json) => {
                    if (!json || !Array.isArray(json)) {
                        throw new Error(`Api error: GET ${path} should return an object array.`);
                    }
                    if (json.length) {
                        return createRecord(json[0]);
                    }
                    return null;
                };
                break;
            case 'count':
                reducer = (json) => {
                    if (!json || typeof json !== 'object' || !json.hasOwnProperty('count')) {
                        throw new Error(`Api error: GET ${path} should return an object with count property.`);
                    }
                    return json.count;
                };
                break;
            case 'findByPk':
            case 'findOne':
            case 'create':
            case 'watch':
                reducer = (json) => json ? createRecord(json) : null;
                break;
            case 'find':
                reducer = (json) => {
                    if (!Array.isArray(json))
                        throw new Error(`Api error: GET ${path} should return an object array.`);
                    return json.map(createRecord);
                };
                break;
            case 'paginate':
                reducer = (json) => {
                    if (json && json.results) {
                        json.results = json.results.map(createRecord);
                    }
                    return json;
                };
                break;
            default:
        }
        this._result = M.request(path, init, this, reducer);
        return this._result;
    }
    _debug() {
        if (debug.enabled) {
            let str = '';
            const M = this.model;
            if (M.name === 'AnonymousModel') {
                str = `Client("${M.path}").`;
            }
            else {
                str = `${M.name}.`;
            }
            str += this._op;
            switch (this._op) {
                case 'findByPk':
                    if (this._id === null) {
                        throw new Error('id is not specified for findByPk');
                    }
                    str += `(${JSON.stringify(this._id)})`;
                    break;
                case 'remove':
                    if (this._id !== null) {
                        str += `(${JSON.stringify(this._id)})`;
                    }
                    else {
                        str += '()';
                    }
                    break;
                case 'update':
                    if (this._id) {
                        str += `(${JSON.stringify(this._id)}, ${JSON.stringify(this._data)})`;
                    }
                    else {
                        str += `(${JSON.stringify(this._data)})`;
                    }
                    break;
                case 'create':
                    str += `(${JSON.stringify(this._data)})`;
                    break;
                default:
                    str += '()';
            }
            if (this._args) {
                let args = this._args;
                str += Object.keys(this._args)
                    .map((key) => `.arg("${key}", ${JSON.stringify(args[key])})`)
                    .join('');
            }
            if (this._filters) {
                str += `.where(${JSON.stringify(this._filters)})`;
            }
            if (this._search) {
                str += `.search("${this._search}")`;
            }
            if (this._limit > 1) {
                str += `.limit(${this._limit})`;
            }
            if (this._page > 1) {
                str += `.page(${this._page})`;
            }
            if (this._sort) {
                str += `.sort(${this._sort})`;
            }
            debug(str);
        }
    }
    _createInit() {
        let init = {};
        init.method = 'GET';
        let path = '';
        let query = {};
        if (this._args) {
            let obj = this._args;
            Object.keys(obj).forEach((key) => {
                query[`_${key}`] = obj[key];
            });
        }
        if (this._filters) {
            Object.assign(query, this._filters);
        }
        if (this._search) {
            query._search = this._search;
        }
        if (this._limit) {
            query._limit = this._limit;
        }
        if (this._page) {
            query._page = this._page;
        }
        if (this._sort) {
            query._sort = this._sort;
        }
        if (this._id !== null && ['findByPk', 'remove', 'update'].indexOf(this._op) > -1) {
            path += `/${encodeURIComponent(this._id)}`;
        }
        if (this._data && ['create', 'update'].indexOf(this._op) > -1) {
            init.body = this._data;
        }
        switch (this._op) {
            case 'count':
                path += '/count';
                break;
            case 'paginate':
                path += '/paginate';
                break;
            case 'watch':
                path += '/watch';
                break;
            case 'create':
                init.method = 'POST';
                break;
            case 'update':
                init.method = 'PATCH';
                break;
            case 'remove':
                init.method = 'DELETE';
                break;
            default:
        }
        init.path = path;
        init.query = query;
        return init;
    }
    then(onSuccess, onFail) {
        return this.exec().then(onSuccess, onFail);
    }
    catch(onFail) {
        return this.exec().catch(onFail);
    }
}
exports.default = Query;
