"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const isBuffer = require("is-buffer");
const Debugger = require("debug");
const stream_1 = require("./stream");
const debug = Debugger('akita:result');
class Result {
    constructor(fetch, path, init, query, reducer) {
        this[Symbol.toStringTag] = 'Promise';
        this._query = query;
        this._reducer = reducer;
        this._path = path;
        this._init = init;
        let promise = fetch(path, init);
        if (debug.enabled) {
            promise = promise.then((res) => {
                debug('response status:', res.status, res.statusText);
                return res;
            }, (error) => {
                debug('fetch error:', error.message);
                return Promise.reject(error);
            });
        }
        this._responsePromise = promise;
    }
    response() {
        return this._responsePromise;
    }
    stream() {
        return this.response().then((res) => res.body);
    }
    ok() {
        return this.response().then((res) => res.ok);
    }
    status() {
        return this.response().then((res) => res.status);
    }
    statusText() {
        return this.response().then((res) => res.statusText);
    }
    size() {
        return this.response().then((res) => res.size);
    }
    headers() {
        return this.response().then((res) => res.headers);
    }
    buffer() {
        return this.response().then((res) => {
            let fn = 'buffer';
            if (res.arrayBuffer) {
                fn = 'arrayBuffer';
            }
            return res[fn]().then((buf) => {
                if (!isBuffer(buf) && buf instanceof ArrayBuffer && typeof Buffer === 'function') {
                    return Buffer.from(buf);
                }
                return buf;
            }, (error) => {
                debug('get buffer error:', error.message);
                return Promise.reject(error);
            });
        });
    }
    blob() {
        return this.response().then((res) => res.blob());
    }
    text() {
        return this.response().then((res) => res.text().then((text) => {
            debug('response text:', text);
            return text;
        }));
    }
    json() {
        return this.response().then((res) => {
            if (res.status === 204) {
                return Promise.resolve();
            }
            return res.text().then((text) => {
                debug('response text:', text);
                if (res.status === 404 && this._query && ['findByPk', 'remove'].indexOf(this._query._op) > -1) {
                    debug(`return null when ${this._query._op} 404`);
                    return null;
                }
                let json;
                try {
                    json = JSON.parse(text);
                }
                catch (error) {
                    return Promise.reject(new Error(`invalid json response body at ${this._path} ${error.message}`));
                }
                if (json && json.error && ['0', 'null', 'none'].indexOf(json.error) < 0) {
                    let error = new Error(json.error);
                    Object.keys(json).forEach((key) => {
                        if (['error', 'message', 'stack'].indexOf(key) === -1) {
                            error[key] = json[key];
                        }
                    });
                    return Promise.reject(error);
                }
                return json;
            }, (error) => {
                debug('json parse error:', error.message);
                return Promise.reject(error);
            });
        });
    }
    then(onSuccess, onFail) {
        if (this._query && this._query._op === 'watch') {
            if (!this._csPromise) {
                this._csPromise = this.stream().then((stream) => new stream_1.default(stream, this._reducer));
            }
            return this._csPromise.then(onSuccess, onFail);
        }
        if (this._reducer) {
            return this.json().then((json) => onSuccess(this._reducer(json)), onFail);
        }
        return this.json().then(onSuccess, onFail);
    }
    catch(onFail) {
        return this.json().catch(onFail);
    }
}
exports.default = Result;
